from pylel.token import Symbols, Token
from .language_core import core
from .language_core.lel_call_function import lel_call_function
from .language_core.standard_language_functions import standard
from .find_base_path import find_base_path

def find_in_scope(scope, name):
	if not scope:
		return None
	if name in scope.variables:
		return scope.variables[name]
	return find_in_scope(scope.upper_scope, name)

def evaluate_expr(scope, expr):
	if type(expr) == list:
		# Evaluate empty block
		if len(expr) == 0:
			return Token(Symbols.LIST, [])

		# List of expressions?
		# Evaluate a block in series
		if type(expr[0]) == list:
			values = list( \
				map(lambda block_expr: evaluate_expr(scope, block_expr), \
					expr))
			return values[-1]
	
		# The rest of the expressions are based on identifiers
		identifier_token = expr[0]
		if identifier_token.type == Symbols.IDENTIFIER:		
			# Core language functions
			if identifier_token.value in core:
				return core[identifier_token.value](evaluate_expr, scope, expr)
			
			# Standard languages functions that manipulate primitives
			if identifier_token.value in standard:
				evaluated_expr = list( \
					map(lambda sub_expr: evaluate_expr(scope, sub_expr),  \
					expr[1:]))
				return standard[identifier_token.value](*evaluated_expr)
			
			# Run a scoped function if one is found
			scoped_function = find_in_scope(scope, identifier_token.value)
			if scoped_function and \
				scoped_function.type == Symbols.FUNCTION_REFERENCE: 
				evaluated_expr = list( \
					map(lambda sub_expr: evaluate_expr(scope, sub_expr),  \
					expr[1:]))
				return lel_call_function(evaluate_expr, scope, \
						evaluated_expr, scoped_function.value)

		# Try and evaluate them single expression
		evaluated = None
		for e in expr:
			evaluated = evaluate_expr(scope, e)
		return evaluated
	else:
		# Return the value of primitives directly in their tokenised form
		if expr.is_token and \
			( \
				expr.type == Symbols.RANGE \
				or expr.type == Symbols.STRING \
				or expr.type == Symbols.NUMBER \
				or expr.type == Symbols.BOOLEAN \
				or expr.type == Symbols.FUNCTION_REFERENCE \
				or expr.type == Symbols.LIST \
			):
			return expr

		# Identifiers will be a function reference or a variable
		if expr.type == Symbols.IDENTIFIER:
			identifier_type = expr.value

			# Pass back variable value. Explicitly check null instead of other
			# falsey values that might really be contained in the variable
			variable_in_scope = find_in_scope(scope, identifier_type)
			if variable_in_scope:
				return variable_in_scope

	raise Exception("Unrecognised expression: {}".format(expr));



from pylel.token import Symbols, Token
from .lel_call_function import lel_call_function

def _get_mapping_function(evaluate_expr, scope, expr):
	def func(l):
		if l.type != Symbols.LIST:
			raise Exception("Invalid list passed to map. Got {}"\
				.format(expr))
		return _perform_mapping(evaluate_expr, scope, expr, l)(evaluate_expr(scope, expr[2]))
	return func

def _perform_mapping(evaluate_expr, scope, expr, l):
	def func(mapping_function):
		if mapping_function.type != Symbols.FUNCTION_REFERENCE:
			raise Exception("Invalid function passed to map. Got {}"\
				.format(expr))
		map_calls = list(\
			map(lambda v: lel_call_function(evaluate_expr, scope, [v[1], \
				Token(Symbols.NUMBER, v[0])], mapping_function.value), enumerate(l.value)))
		return Token(Symbols.LIST, map_calls)
	return func

def lel_map(evaluate_expr, scope, expr):
	return _get_mapping_function(evaluate_expr, scope, expr)(evaluate_expr(scope, expr[1]))




class Scope(object):
	def __init__(self, upper_scope, base_path = None):
		self.upper_scope = upper_scope
		self.variables = {}
		self.base_path = base_path

	def __repr__(self):
		return str(self.__dict__) + "\n" 

	def __str__(self):
		return self.__repr__()



from .evaluate_expr import evaluate_expr
from pylel.tools import get_real_dir_name
from .scope import Scope

def evaluate(ast, base_path = None):
	if not base_path:
		base_path = get_real_dir_name(__file__)
	root_scope = Scope(None, base_path)
	ast_evaluated = list( \
		map(lambda expr: evaluate_expr(root_scope, expr), ast))
	return ast_evaluated



import math, functools
from pylel.token import Symbols, Token

def _addition(*numbers):
	if any([number.type != Symbols.NUMBER for number in numbers]):
		raise Exception("+ only operates on NUMBER type")
	return functools.reduce(lambda acc, cur: Token(Symbols.NUMBER, acc.value + cur.value), numbers)

def _subtraction(x, y):
	if x.type != Symbols.NUMBER or y.type != Symbols.NUMBER:
		raise Exception("- only operates on NUMBER type")
	return Token(Symbols.NUMBER, x.value - y.value)

def _division(x, y):
	if x.type != Symbols.NUMBER or y.type != Symbols.NUMBER:
		raise Exception("/ only operates on NUMBER type")
	if y.value == 0:
		raise Exception("Divison by zero error")
	return Token(Symbols.NUMBER, x.value / y.value)

def _multiplication(*numbers):
	if any([number.type != Symbols.NUMBER for number in numbers]):
		raise Exception("* only operates on NUMBER type")
	return functools.reduce(lambda acc, cur: Token(Symbols.NUMBER, acc.value * cur.value), numbers)

def _power(x, y):
	if x.type != Symbols.NUMBER or y.type != Symbols.NUMBER:
		raise Exception("** only operates on NUMBER type")
	return Token(Symbols.NUMBER, pow(x.value, y.value))

def _mod(x, y):
	if x.type != Symbols.NUMBER or y.type != Symbols.NUMBER:
		raise Exception("\% only operates on NUMBER type")
	if y.value == 0:
		raise Exception("Divison by zero error")
	return Token(Symbols.NUMBER, x.value % y.value)

def _sin(x):
	if x.type != Symbols.NUMBER:
		raise Exception("sin only operates on NUMBER type")
	return Token(Symbols.NUMBER, math.sin(x.value))

lel_math = {
	"+": _addition,
	"-": _subtraction,
	"*": _multiplication,
	"/": _division,
	"**": _power,
	"%": _mod,
	"sin": _sin 
}



import os

def read_file(file_name):
	read_data = None
	try:
		with open(file_name, "r+b") as f:
			read_data = f.read().decode("utf-8") 
	except:
		raise Exception("Couldn't read the file. Please check the filename and the path.")
	return read_data

def get_real_dir_name(file):
	return os.path.dirname(os.path.realpath(file))

def is_int(a):
	return a % 1 == 0



from pylel.token import Symbols

def _unescape_characters(str):
	return str.replace(r'\n', '\n') \
	.replace(r'\r', '\r') \
	.replace(r'\f', '\f') \
	.replace(r'\b', '\b') \
	.replace(r'\t', '\t') \
	.replace(r'\v', '\v') \
	.replace(r'\;', ';')

def _clean_string(str):
	return _unescape_characters(str[1:-1])

def _clean_bool(bool):
	return bool == "true"

def _clean_number(number):
	return float(number)

def _clean_token(token):
	if token.type == Symbols.STRING:
		token.value = _clean_string(token.value)
	elif token.type == Symbols.NUMBER:
		token.value = _clean_number(token.value)
	elif token.type == Symbols.BOOLEAN:
		token.value = _clean_bool(token.value)
	return token

class Parse(object):
	depth_pointer = 0
	
	@staticmethod
	def _add_token_to_expr_tree(ast, token):
		level = ast
		for i in range(Parse.depth_pointer):
			level = level[len(level) - 1]
		level.append(token)
	
	@staticmethod
	def _pop_expr():
		Parse.depth_pointer -= 1

	@staticmethod
	def _push_expr(ast):
		Parse._add_token_to_expr_tree(ast, [])
		Parse.depth_pointer += 1

	@staticmethod
	def parse(tokens):
		Parse.depth_pointer = 0
		clean_tokens = [_clean_token(token) for token in tokens]
		ast = []
		for token in clean_tokens:
			if token.type == Symbols.LPAREN:
				Parse._push_expr(ast)
			elif token.type == Symbols.RPAREN:
				Parse._pop_expr()
			else:
				Parse._add_token_to_expr_tree(ast, token)
		return ast



from pylel.token import Symbols, Token

EMPTY_LIST = Token(Symbols.LIST, [])

def _join(l):
	if l.type != Symbols.LIST:
		raise Exception("Can't join non-lists. Got {}".format(l.type))
	return Token(Symbols.STRING, "".join(list(map(\
		lambda list_item: list_item.value, l.value))))

def _length(l):
	if l.type != Symbols.LIST:
		raise Exception("length operates on LIST type. Got {}".format(l.type))
	return Token(Symbols.NUMBER, len(l.value))

def _head(l):
	if l.type != Symbols.LIST:
		raise Exception("head operates on LIST type. Got {}".format(l.type))
	return l.value[0] if len(l.value) != 0 else EMPTY_LIST

def _tail(l):
	if l.type != Symbols.LIST:
		raise Exception("tail operates on LIST type. Got {}".format(l.type))
	return Token(Symbols.LIST, l.value[1:])

def _nth(l, n):
	if l.type != Symbols.LIST:
		raise Exception("nth operates on LIST type. Got {}".format(l.type))
	nth = int(n.value)
	if nth >= 0 and nth < len(l.value):
		return l.value[nth]
	raise Exception("nth: bad index {}. Given list has {} elements"\
		.format(n.value, len(l.value)))

def _sublist(l, start, end):
	if l.type != Symbols.LIST:
		raise Exception("sublist operates on LIST type. Got {}".format(l.type))
	list_length = len(l.value)
	s = int(start.value)
	e = int(end.value)

	if s > e:
		raise Exception("start index cannot be greater than the end index for a sublist")
	if s < 0 or e >= list_length:
		raise Exception("sublist indexes out of range. Got start ({}) end ({}) for list of length {}"\
			.format(s, e, list_length))
	return Token(Symbols.LIST, l.value[s:e + 1])

lel_list = {
	"join": _join,
	"length": _length,
	"head": _head,
	"tail": _tail,
	"nth": _nth,
	"sublist": _sublist
}



from pylel.token import Symbols
from pylel.evaluate.scope import Scope

def lel_call_function(evaluate_expr, scope, args, function_descriptor):
	# Every time the function runs it gets it's own scope, 
	# meaning variables set inside this function
	# will not persist across different calls.
	execution_scope = Scope(function_descriptor.scope)
	if len(args) != len(function_descriptor.expected_arguments):
		raise Exception("Expected {} arguments for function {} but got {}"\
			.format(len(function_descriptor.expected_arguments), \
				function_descriptor.name, len(args)))

	# Place arguments into the execution scope
	for i, argument in enumerate(args):
		execution_scope.variables[function_descriptor.expected_arguments[i]] = \
		argument
	body_evaluators = list(\
		map(lambda f_expr: evaluate_expr(execution_scope, f_expr), \
			function_descriptor.body_expressions))
	return body_evaluators[-1]



from pylel.token import Symbols
from .lel_call_function import lel_call_function
from .lel_lambda import lel_lambda

def _perform_function_call(evaluate_expr, scope, expr, function_descriptor):
	def func(args):
		return lel_call_function(evaluate_expr, scope, args, function_descriptor)
	return func

def _get_function_args(evaluate_expr, scope, expr):
	def func(f_reference):
		if f_reference and f_reference.is_token and f_reference.type == Symbols.FUNCTION_REFERENCE:
			function_descriptor = f_reference.value
			if len(expr) >= 3:
				expressions = list(\
					map(lambda sub_expr: evaluate_expr(scope, sub_expr), expr[2:]))
				return _perform_function_call(evaluate_expr, scope, expr, function_descriptor)\
				(expressions)
			else:
				return _perform_function_call(evaluate_expr, scope, expr, function_descriptor)([])
		else:
			raise Exception("First argument must be a FUNCTION_REFERENCE. Got {} for expression ${}"\
				.format(expr[1].type, expr))
	return func

def lel_call(evaluate_expr, scope, expr):
	if len(expr) >= 2:
		if type(expr[1]) == list or \
		(expr[1].is_token and expr[1].type == Symbols.IDENTIFIER):
			return _get_function_args(evaluate_expr, scope, expr)(evaluate_expr(scope, expr[1]))
	raise Exception("{} is not a valid function reference"\
		.format(expr[1]))
	




# -*- coding: utf-8 -*-
from __future__ import print_function
import re
import six
from pylel.tools import get_real_dir_name
from pylel.token import tokenise, Symbols
from pylel.parse import Parse
from pylel.evaluate.scope import Scope
from pylel.evaluate.evaluate_expr import evaluate_expr

LPAR = "("
RPAR = ")"
REPL_LINE = "> "

def parantheses_balance(str):
	left = len(re.findall("\(", str))
	right = len(re.findall("\)", str))
	return 0 if left == right else -1 if left > right else 1

def repl_execute(expr_str, root_scope):
	tokens = tokenise(expr_str)
	ast = Parse.parse(tokens)
	evaluate_expr(root_scope, ast[0])
	print("")

def _repl(root_scope, strikes = 0):
	expr = ""
	try:
		while True:
			line = six.moves.input()
			if len(line) == 0:
				print("> ", end = '')
				continue
			expr += line
			balance = parantheses_balance(expr)
			if balance == 0:
				if expr[0] != LPAR:
					expr = LPAR + expr + RPAR
				strikes = 0
				repl_execute(expr, root_scope)
				expr = ""
			elif balance == -1:
				expr += " "
			else:
				print("Too many ')'!")
				expr = ""
			print(REPL_LINE, end = '')
	except KeyboardInterrupt:
		print("\n")
		if strikes == 0:
			print(REPL_LINE, end = '')
			_repl(root_scope, strikes + 1)
	except Exception as e:
		print(e)
		print(REPL_LINE, end = '')
		_repl(root_scope)

def repl():
	print("Lel REPL - Ömer Saatcioglu 2017\n(Original Node.js version by Francis Stokes)\n> ", end = '')
	root_scope = Scope(None, get_real_dir_name(__file__))
	_repl(root_scope)



from pylel.token import Symbols, Token
from .lel_call_function import lel_call_function

def _perform_filtering(evaluate_expr, scope, expr, l):
	def func(filtering_function):
		if filtering_function.type != Symbols.FUNCTION_REFERENCE:
			raise Exception("Invalid function passed to filter. Got {}"\
				.format(expr))

		new_list = [t for i, t in enumerate(l.value)\
			if lel_call_function(evaluate_expr, scope,\
				[t, Token(Symbols.NUMBER, i)], filtering_function.value).value]

		return Token(Symbols.LIST, new_list)
	return func

def _get_filtering_function(evaluate_expr, scope, expr):
	def func(l):
		if l.type != Symbols.LIST:
			raise Exception("Invalid list passed to filter. Got {}"\
				.format(expr))
		evaluted_expr = evaluate_expr(scope, expr[2])
		return _perform_filtering(evaluate_expr, scope, \
			expr, l)(evaluted_expr)
	return func

def lel_filter(evaluate_expr, scope, expr):
	return _get_filtering_function(evaluate_expr, scope, expr)\
	(evaluate_expr(scope, expr[1]))




def lel_if(evaluate_expr, scope, expr):
	if len(expr) < 4:
		raise Exception("Conditional expressions require 3 arguments. Got {} at expression {}"\
			.format(len(expr), expr))

	conditional_result = evaluate_expr(scope, expr[1])
	branch = expr[2] if conditional_result.value else expr[3]
	return evaluate_expr(scope, branch)




from pylel.token import Symbols, Token
from .lel_call_function import lel_call_function
from .lel_lambda import lel_lambda

def _perform_function_call(evaluate_expr, scope, expr, function_descriptor):
	def func(args):
		if args.type != Symbols.LIST:
			raise Exception("Referenced arguments must be a LIST apply ({}). Got {} for expression {}"\
				.format(function_descriptor.name, args.type, expr))
		return lel_call_function(evaluate_expr, scope, args.value, function_descriptor)
	return func

def _get_function_args(evaluate_expr, scope, expr):
	def func(f_reference):
		if f_reference and f_reference.is_token and f_reference.type == Symbols.FUNCTION_REFERENCE:
			function_descriptor = f_reference.value
			if len(expr) >= 3:
				return _perform_function_call(evaluate_expr, scope, expr, function_descriptor)\
				(evaluate_expr(scope, expr[2]))
			else:
				empty_list = Token(Symbols.LIST, [])
				return _perform_function_call(evaluate_expr, scope, expr, function_descriptor)(empty_list)
		else:
			raise Exception("First argument must be a FUNCTION_REFERENCE. Got {} for expression ${}"\
				.format(expr[1].type, expr))
	return func

def lel_apply(evaluate_expr, scope, expr):
	if len(expr) >= 2:
		f_ref_is_array = type(expr[1]) == list
		f_ref_is_identifier = expr[1].is_token and expr[1].type == Symbols.IDENTIFIER
		if f_ref_is_array or f_ref_is_identifier:
			return _get_function_args(evaluate_expr, scope, expr)(evaluate_expr(scope, expr[1]))
	raise Exception("{} is not a valid function reference"\
	.format(expr[1]))




from .lel_boolean import lel_boolean
from .lel_general import lel_general
from .lel_list import lel_list
from .lel_math import lel_math
from .lel_string import lel_string

standard = lel_general.copy()
standard.update(lel_boolean)
standard.update(lel_math)
standard.update(lel_list)
standard.update(lel_string)



from .lel_if import lel_if
from .lel_let import lel_let
from .lel_lambda import lel_lambda
from .lel_list import lel_list
from .lel_map import lel_map
from .lel_filter import lel_filter
from .lel_call import lel_call
from .lel_apply import lel_apply
from .lel_mutate import lel_mutate
from .lel_import import lel_import
from .lel_function import lel_function

core = {
	"if": lel_if,
	"let": lel_let,
    "function": lel_function,
    "lambda": lel_lambda,
    "list": lel_list,
    "map": lel_map,
    "filter": lel_filter,
    "call": lel_call,
    "apply": lel_apply,
    "mutate": lel_mutate,
    "import": lel_import
}



from pylel.token import Symbols

def lel_mutate(evaluate_expr, scope, expr):
  if expr[1].type != Symbols.IDENTIFIER:
    raise Exception("Variable name must be an IDENTIFER. Got {} for {}"\
      .format(expr[1].type, expr[1].value))
  name = expr[1].value
  if not name in scope.variables:
    raise Exception("No variable '{}' to mutate in the local scope"\
      .format(name))
  value = evaluate_expr(scope, expr[2])
  scope.variables[name] = value
  return value



from pylel.token import Symbols, Token
from pylel.evaluate.scope import Scope
from .lel_function import ScopedFunction

def _create_expected_arguments(token):
	if token.type != Symbols.IDENTIFIER:
		raise Exception()
	return token.value

def lel_lambda(evaluate_expr, scope, expr):
	f_name = "lambda_function"
	try:
		expected_arguments = list(map(_create_expected_arguments, expr[1]))
	except:
		raise Exception("Function declaration arguments must be an IDENTIFIER. Got {} for function {}"\
			.format(token.type, f_name))
	f_body = expr[2:]
	if len(f_body) == 0:
		raise Exception("Function body must contain at least one statement. Got none for function {}"\
			.format(f_name))
	function_scope = Scope(scope)
	return Token(Symbols.FUNCTION_REFERENCE, \
		ScopedFunction(f_name, f_body, expected_arguments, function_scope))



from pylel.token import Symbols, Token

def _split(string):
	if string.type != Symbols.STRING:
		raise Exception("Can't split non-strings. Got {}".format(string.type))
	return Token(Symbols.LIST, [Token(Symbols.STRING, c) for c in string.value])

lel_string = {
	"split": _split
}



from pylel.token import Symbols, Token
from pylel.evaluate.scope import Scope

class ScopedFunction(object):
	def __init__(self, name, body_expressions = [], expected_arguments = [], scope = {}):		
		self.name = name
		self.is_function = True
		self.body_expressions = body_expressions
		self.expected_arguments = expected_arguments
		self.scope = scope

	def __repr__(self):
		return str(self.__dict__) + "\n" 

	def __str__(self):
		return self.__repr__()

def _create_expected_args(token):
	if token.type != Symbols.IDENTIFIER:
		raise Exception()
	return token.value

def lel_function(evaluate_expr, scope, expr):
	if expr[1].type != Symbols.IDENTIFIER:
		raise Exception("Function name must be an IDENTIFIER. Got {} for {}"\
			.format(expr[1].type, expr[1].value))

	f_name = expr[1].value
	try:
		expected_arguments = list( \
			map(lambda token: _create_expected_args(token), \
				expr[2]
				))
	except:
		raise Exception("Function declaration arguments must be an IDENTIFIER. Got {} for function {}"\
			.format(token.type, f_name))

	f_body = expr[3:]
	if len(f_body) < 1:
		raise Exception("Function body must contain at least one statement. Got none for function {}"\
			.format(f_name))
	
	function_scope = Scope(scope)
	scope.variables[f_name] = Token( \
		Symbols.FUNCTION_REFERENCE, \
		ScopedFunction(f_name, f_body, expected_arguments, function_scope)
		)
	return scope.variables[f_name]



from pylel.token import tokenise, Symbols
from pylel.tools import read_file, get_real_dir_name
from pylel.parse import Parse
from pylel.evaluate import evaluate

def validate(tokens):
	left = 0
	right = 0
	for token in tokens:
		if token.type == Symbols.LPAREN: left += 1 
		if token.type == Symbols.RPAREN: right += 1 
		if right > left:
			return False
	return left == right

def interpreter(file_name):
	data = read_file(file_name)
	tokens = tokenise(data)
	if not validate(tokens):
		raise Exception('Unmatched parentheses.')
	ast = Parse.parse(tokens)
	base_path = get_real_dir_name(file_name)
	return evaluate(ast, base_path)



from os import path
from pylel.token import Symbols
from pylel.evaluate.scope import Scope
from pylel.evaluate.evaluate_expr.find_base_path import find_base_path
from pylel.token import tokenise
from pylel.tools import read_file, get_real_dir_name

def inject_module_to_scope(scope, module_scope, file_path):
	intersect = set(scope.variables.keys()) & set(module_scope.variables.keys())
	if len(intersect) != 0:
		raise Exception("Cannot overwrite variable in scope {} from module {}"\
			.format(intersect, file_path))
	scope.variables.update(module_scope.variables)

def lel_import(evaluate_expr, scope, expr):
	from interpreter import interpreter
	filename = evaluate_expr(scope, expr[1])
	if filename.type != Symbols.STRING:
		raise Exception("Import path must resolve to a string. Got {}"\
			.format(filename))
	file_path = path.join(find_base_path(scope), filename.value)
	base_path = get_real_dir_name(file_path)
	ast = interpreter(file_path)
	module_scope = Scope(None, base_path)
	evaluated = evaluate_expr(module_scope, ast)
	inject_module_to_scope(scope, module_scope, file_path)
	return evaluated




# -*- coding: utf-8 -*-
import sys

from pylel.repl import repl
from pylel.interpreter import interpreter

help_text = "LEL Language Interpreter in Python by Ömer Saatcioglu 2017\n(Original Node.js version by Francis Stokes)\nhttps://github.com/osaatcioglu/Lisp-esque-language\nUsage:\npylel [filename] [filename] ..."

def main(filename = None):
	sys.setrecursionlimit(0x100000)
	try:
		if filename:
			interpreter(filename)
		else:
			repl()
	except Exception as e:
		print(str(e))
		sys.exit(1)

def package_main():
	if len(sys.argv) == 1:
		return main()
	if "--" in sys.argv[1]:
		print(help_text)
		return
	for filename in sys.argv[1:]:
		main(filename)

if __name__ == "__main__":
	package_main()



import re

class Symbols(object):
	# Whitespace skips
	SKIP = 'SKIP'

	# List operator expanded in parser
	RANGE = 'RANGE'

	# Expression start/end
	LPAREN = 'LPAREN'
	RPAREN = 'RPAREN'

	# Primitives
	NUMBER = 'NUMBER'
	STRING = 'STRING'
	BOOLEAN = 'BOOLEAN'
	IDENTIFIER = 'IDENTIFIER'

	# Used internally for passing functions
	FUNCTION_REFERENCE = 'FUNCTION_REFERENCE'

	# Used internally for describing lists
	LIST = 'LIST'

number = ['^\-?[0-9]+\.?[0-9]*\Z', Symbols.NUMBER]
string = ['^\"[^\n\"]*\"\Z', Symbols.STRING]
whitespace = ['^[\s\n]+\Z', Symbols.SKIP]
comment = ['^;.+?\n\Z', Symbols.SKIP]
identifier = ['^[a-zA-Z\+\-\/\*\%\_\>\<=]*\Z', Symbols.IDENTIFIER]
boolTrue = ['true', Symbols.BOOLEAN]
boolFalse = ['false', Symbols.BOOLEAN]
lparen = ['(', Symbols.LPAREN]
rparen = [')', Symbols.RPAREN]
lel_range = ['..', Symbols.RANGE]

boolean_not_end = "^[^ \n\(\)]\Z"

class Patterns(object):
	ambiguous = [
	  ['^\-\Z', [number]]
	]
	exact = [
	  boolTrue,
	  boolFalse,
	  lparen,
	  rparen,
	  lel_range
	]
	tokens = [
	  whitespace,
	  comment,
	  number,
	  string,
	  identifier
	]

class Token(object):
	def __init__(self, t, value):
		self.type = t
		self.value = value
		self.is_token = True

	def to_string(self):
		return str(self.type)

	def __repr__(self):
		return str(self.__dict__) + "\n" 

	def __str__(self):
		return self.__repr__()

def tokenise(chars):
	tokens = []
	check = ""
	i = 0
	while i < len(chars):
		check += chars[i]

		# Check against exact patterns first
		if len(check) == 1:
			matched_exact_pattern = False
			for exact_str, symbol in Patterns.exact:
				if len(exact_str) + i <= len(chars):
					exact_check = check + chars[i + 1: i + len(exact_str)]
					if exact_check == exact_str:
						# Boolean should be whole word only 
						# since it can be used within a variable name
						if symbol == Symbols.BOOLEAN and \
							re.match(boolean_not_end, chars[i + len(exact_str)]):
							break
						# Set the new i pointer
						i += len(exact_str) - 1

						# Add the token to the list
						if symbol != Symbols.SKIP:
							tokens.append(Token(symbol, exact_check))
						# Reset the check string
						check = ""

						# Exit from the token search
						matched_exact_pattern = True
						break
			if matched_exact_pattern:
				i += 1
				continue

		# Perform an ambiguous check to prioritise a pattern match
		if len(check) == 1 and i < len(chars) - 1:
			for re_str, patterns in Patterns.ambiguous:
				if re.match(re_str, check):
					for re_str_ambiguous, _ in patterns:
						re_p_ambiguous = re.compile(re_str_ambiguous)
						if re_p_ambiguous.match(check) or re_p_ambiguous.match(check + chars[i + 1]):
							i += 1
							check += chars[i]
							break

		# Test for tokens until a sucessful one is found
		for re_str, label in Patterns.tokens:
			re_p = re.compile(re_str)
			is_match_peek_check = False
			if re_p.match(check):
				if i == len(chars) - 1:
					# Add the token to the list
					if label != Symbols.SKIP:
						tokens.append(Token(label, check))
					break
				# Peek ahead at the next charcters while it's still matching the same token
				peek_check = check
				for j in range(i + 1, len(chars)):
					peek_check += chars[j]
					# Does checking with another character still match?
					if not re_p.match(peek_check):
						# If not, consider everything up until the last peekCheck the token
						check = peek_check[:-1]
						
						# Set tokeniser index to this point
						i = j - 1
						# Add the token to the list
						if check and label != Symbols.SKIP:
							tokens.append(Token(label, check))
						check = ""
						# Exit from the token search
						is_match_peek_check = True
						break
			if is_match_peek_check:
				break

		i += 1
	return tokens



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='lel-lang',

    version='0.0.0b5',

    description='Lel programming language(Lisp-esque language)',
    long_description=long_description,

    url='https://github.com/osaatcioglu/Lisp-esque-language',

    author='Ömer Saatcioglu',
    author_email='osaatcioglu@gmail.com',

    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Topic :: Software Development :: Compilers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],

    keywords='compiler lisp lel parser ast tokenize',

    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    install_requires=['six'],

    extras_require={
        'dev': [],
        'test': [],
    },

    package_data={},

    entry_points={
        'console_scripts': [
            'pylel=pylel:package_main',
        ],
    },
)



from pylel.token import Symbols, Token

lel_boolean = {
	'=': lambda x, y: Token(Symbols.BOOLEAN, x.value == y.value ),
	'<': lambda x, y: Token(Symbols.BOOLEAN, x.value < y.value ),
	'>': lambda x, y: Token(Symbols.BOOLEAN, x.value > y.value ),
	'<=': lambda x, y: Token(Symbols.BOOLEAN, x.value <= y.value ),
	'>=': lambda x, y: Token(Symbols.BOOLEAN, x.value >= y.value ),
}



from pylel.token import Symbols

def lel_let(evaluate_expr, scope, expr):
	if expr[1].type != Symbols.IDENTIFIER:
		raise Exception("Variable name must be an IDENTIFER. Got {} for {}"\
			.format(expr[1].type, expr[1].value))
	name = expr[1].value
	if name in scope.variables:
		raise Exception("Can't implicitly mutate previously assigned scoped variable '{}'"\
			.format(name))
	value = evaluate_expr(scope, expr[2])
	scope.variables[name] = value
	return value



from __future__ import print_function
import sys, functools
from pylel.token import Symbols, Token
from pylel.tools import is_int

EMPTY_LIST = Token(Symbols.LIST, [])

def _pretty_string(token):
	if token.type in [Symbols.NUMBER, Symbols.BOOLEAN, Symbols.STRING]:
		# Not all numbers are float. Fixing it before printing
		if token.type == Symbols.NUMBER:
			return str(int(token.value) if is_int(token.value) else token.value)
		return str(token.value)
	elif token.type == Symbols.LIST:
		return "(" + ", ".join(list(map(_pretty_string, token.value))) + ")"
	return str(token)

def _exit(code_token):
	code = code_token.value if (code_token and code_token.type == Symbols.NUMBER) else 0
	sys.exit(code)

def _print(*items):
	out = "".join(list(map(_pretty_string, items)))
	print(out, end='')
	return Token(Symbols.STRING, out)

def _cls():
	print("\033c")
	return EMPTY_LIST

def _concat(*concatables):
	if len(concatables) == 0:
		return EMPTY_LIST
	t = concatables[0].type
	all_same_type = all([concatable.type == t for concatable in concatables])

	if t == Symbols.LIST and all_same_type:
		return Token(Symbols.LIST, [item \
			for sublist in concatables for item in sublist.value])
	else:
		return functools.reduce(lambda acc, cur: Token(Symbols.STRING, acc.value + cur.value \
			if cur.type in [Symbols.STRING, Symbols.NUMBER, Symbols.BOOLEAN] \
			else str(cur)), concatables)

lel_general = {
	"exit": _exit,
	"print": _print,
	"cls": _cls,
	"concat": _concat,
}








from pylel.token import Symbols, Token

def _expand_ranges(values):
	found_range = any(value.type == Symbols.RANGE for value in values)
	if not found_range:
		return Token(Symbols.LIST, values)

	if len(values) != 3:
		raise Exception("List with range operator requires exactly 3 arguments")
	prev = values[0]
	next_ = values[2]
	if prev.type != Symbols.NUMBER or next_.type != Symbols.NUMBER:
		raise Exception("Cannot make range from non-number")

	range_direction = prev.value < next_.value
	start = prev.value if range_direction else next_.value
	end = next_.value if range_direction else prev.value
	expanded_range = []

	for j in range(int(start), int(end) + 1):
		expanded_range.append(Token(Symbols.NUMBER, j))
	if not range_direction:
		expanded_range = expanded_range[::-1]
	return Token(Symbols.LIST, expanded_range)

def lel_list(evaluate_expr, scope, expr):
	expressions = list(\
		map(lambda sub_expr: evaluate_expr(scope, sub_expr), expr[1:]))
	return _expand_ranges(expressions)
